@@ -11,7 +11,7 @@
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
 
 class spider_db_mbase_util: public spider_db_util
 {
@@ -29,6 +29,17 @@ public:
     uint name_length,
     CHARSET_INFO *name_charset
   );
+  int append_escaped_name(
+    spider_string *str,
+    const char *name,
+    uint name_length
+  );
+  int append_escaped_name_with_charset(
+    spider_string *str,
+    const char *name,
+    uint name_length,
+    CHARSET_INFO *name_charset
+  );
   bool is_name_quote(
     const char head_code
   );
@@ -64,10 +75,26 @@ public:
     spider_string *str,
     bool sql_log_off
   );
+  int append_wait_timeout(
+    spider_string *str,
+    int wait_timeout
+  );
+  virtual int append_sql_mode_internal(
+    spider_string *str,
+    sql_mode_t sql_mode
+  );
+  int append_sql_mode(
+    spider_string *str,
+    sql_mode_t sql_mode
+  );
   int append_time_zone(
     spider_string *str,
     Time_zone *time_zone
   );
+  int append_loop_check(
+    spider_string *str,
+    SPIDER_CONN *conn
+  );
   int append_start_transaction(
     spider_string *str
   );
@@ -172,6 +199,7 @@ public:
     spider_string *str
   );
 #endif
+  bool append_charset_name_before_string();
 };
 
 class spider_db_mysql_util: public spider_db_mbase_util
@@ -179,6 +207,13 @@ class spider_db_mysql_util: public spider_db_mbase_util
 public:
   spider_db_mysql_util();
   ~spider_db_mysql_util();
+  int append_column_value(
+    ha_spider *spider,
+    spider_string *str,
+    Field *field,
+    const uchar *new_ptr,
+    CHARSET_INFO *access_charset
+  );
 };
 
 class spider_db_mariadb_util: public spider_db_mbase_util
@@ -186,6 +221,17 @@ class spider_db_mariadb_util: public spider_db_mbase_util
 public:
   spider_db_mariadb_util();
   ~spider_db_mariadb_util();
+  int append_sql_mode_internal(
+    spider_string *str,
+    sql_mode_t sql_mode
+  );
+  int append_column_value(
+    ha_spider *spider,
+    spider_string *str,
+    Field *field,
+    const uchar *new_ptr,
+    CHARSET_INFO *access_charset
+  );
 };
 
 class spider_db_mbase_row: public spider_db_row
@@ -267,20 +313,22 @@ public:
   );
   int fetch_table_status(
     int mode,
-    ha_rows &records,
-    ulong &mean_rec_length,
-    ulonglong &data_file_length,
-    ulonglong &max_data_file_length,
-    ulonglong &index_file_length,
-    ulonglong &auto_increment_value,
-    time_t &create_time,
-    time_t &update_time,
-    time_t &check_time
+    ha_statistics &stat
+  );
+  int fetch_simple_action(
+    uint simple_action,
+    uint position,
+    void *param
   );
   int fetch_table_records(
     int mode,
     ha_rows &records
   );
+#ifdef HA_HAS_CHECKSUM_EXTENDED
+  int fetch_table_checksum(
+    ha_spider *spider
+  );
+#endif
   int fetch_table_cardinality(
     int mode,
     TABLE *table,
@@ -396,7 +444,7 @@ public:
   bool is_xa_nota_error(
     int error_num
   );
-  void print_warnings(
+  int print_warnings(
     struct tm *l_time
   );
   spider_db_result *store_result(
@@ -405,11 +453,17 @@ public:
     int *error_num
   );
   spider_db_result *use_result(
+    ha_spider *spider,
     st_spider_db_request_key *request_key,
     int *error_num
   );
   int next_result();
   uint affected_rows();
+  uint matched_rows();
+  bool inserted_info(
+    spider_db_handler *handler,
+    ha_copy_info *copy_info
+  );
   ulonglong last_insert_id();
   int set_character_set(
     const char *csname
@@ -466,11 +520,26 @@ public:
     bool sql_log_off,
     int *need_mon
   );
+  bool set_wait_timeout_in_bulk_sql();
+  int set_wait_timeout(
+    int wait_timeout,
+    int *need_mon
+  );
+  bool set_sql_mode_in_bulk_sql();
+  int set_sql_mode(
+    sql_mode_t sql_mode,
+    int *need_mon
+  );
   bool set_time_zone_in_bulk_sql();
   int set_time_zone(
     Time_zone *time_zone,
     int *need_mon
   );
+  bool set_loop_check_in_bulk_sql();
+  int set_loop_check(
+    int *need_mon
+  );
+  int fin_loop_check();
   int exec_simple_sql_with_result(
     SPIDER_TRX *trx,
     SPIDER_SHARE *share,
@@ -660,6 +729,9 @@ public:
     spider_string *str
   );
 #endif
+#ifdef HA_HAS_CHECKSUM_EXTENDED
+  bool checksum_support();
+#endif
 protected:
   int create_table_names_str();
   void free_table_names_str();
@@ -745,6 +817,7 @@ public:
   spider_mbase_share      *mysql_share;
   SPIDER_LINK_FOR_HASH    *link_for_hash;
   uchar                   *minimum_select_bitmap;
+  uchar                   direct_insert_kind;
   spider_mbase_handler(
     ha_spider *spider,
     spider_mbase_share *share,
@@ -1319,15 +1392,15 @@ public:
     int link_idx
   );
   int append_explain_select_part(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
   int append_explain_select(
     spider_string *str,
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
@@ -1439,16 +1512,25 @@ public:
     int link_idx,
     int crd_mode
   );
+  int simple_action(
+    uint simple_action,
+    int link_idx
+  );
   int show_records(
     int link_idx
   );
+#ifdef HA_HAS_CHECKSUM_EXTENDED
+  int checksum_table(
+    int link_idx
+  );
+#endif
   int show_last_insert_id(
     int link_idx,
     ulonglong &last_insert_id
   );
   ha_rows explain_select(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     int link_idx
   );
   int lock_tables(
@@ -1592,6 +1674,18 @@ public:
     spider_fields *fields
   );
 #endif
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  bool check_direct_update(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+  bool check_direct_delete(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+#endif
 };
 
 class spider_mysql_handler: public spider_mbase_handler
